#!/usr/bin/python
##############################################################################
#
# Copyright (c) 2002, 2006 Nexedi SARL and Contributors. All Rights Reserved.
#                    Bartlomiej Gorny <bartek@erp5.pl>
#                    Jacek Medrzycki <jacek@erp5.pl>
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsability of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# garantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################


import time

import config
import factory
import start
from logger import Log

class Pool(object):
  """
    manages a pool of worker object; number of workers is
    set in config directs requests to the first available
    worker uses semaphore to hold request untill one is
    available 
  """
  total = config.pool_size
  inst = {}
  timers = {}

  def __init__(self):
    """
      set up semaphore, make workers
    """
    #self.initializePool()

  def getWorkerObject(self, worker=None):
    """
      New way of getting worker object
      if worker is None, only check if one is available
      otherwise return false, all the rest is done by dispatcher
      if worker is given as an argument, this method kills it
      (possibly tests it first, maybe it is good?) - but if it is frozen, 
      the only way to test is time out...
      and produces and returns a new one
      XXX it should produce a new one first, and then replace the old one
      or kill if new one could not be made
      so that getSize always returns the right value
    """
    if worker is None:
      for w in self.inst.values():
        if w and not w.busy:
          w.busy = True
          return w
      return False
    # else we produce and return a new worker (flagged as busy, to maintain consistency)
    # we assume we are getting a worker to do something with it
    i = worker.idx
    self.makeWorkerObject(i)
    self.inst[i].busy = True
    return self.inst[i]

  def releaseWorkerObject(self, worker):
    """
      Release worker so that it is available
      Return True if successful (successful means that the worker was busy and we set it free)
      If something went wrong, we kill the worker
    """
    try:
      if not worker.busy:
        return False
      worker.busy = False
    except Exception, e:
      Log.logException(e)
      if hasattr(worker, 'idx'): # sometimes it can be None, for example
        self.inst[worker.idx] = None
      return False
    return True

  def getSize(self):
    """
      Impl. 2
      return number of existing workers (busy or not)
      so that we can check if the pool is usable at all
    """
    return len([w for w in self.inst.values() if w])

  def initializePool(self,offset=0):
    """
      Impl. 2
      kill all workers if any are present
      build new ones
    """
    self.offset=offset
    for w in self.inst.values():
      self.releaseWorkerObject(w)
    self.inst = {}
    #for i in range(self.total):
    #now we create only ONE worker
    self.makeWorkerObject(1)

  def makeWorkerObject(self, i):
    """
      Impl. 2
      make worker object and OOo instance for it (no need to run start.py separately)
      kill OOo if existed
      XXX this method needs refactoring
    """
    try:
      self.inst[i] = None
      # start OOo
      start.killInstance(self.offset)
      start.startInstance(self.offset, keep_control=True)
      time.sleep(config.instance_load_time)
      try:
        # build worker
        self.inst[i] = (factory.builder.build(self, self.offset))  #XXX now we give offset instead of w. number
      except Exception, e:
        if getattr(e, 'typeName', None) == 'com.sun.star.connection.NoConnectException':
          # no connection
          Log.debug('no connection - will retry in %d' % (config.instance_load_time * 2))
          time.sleep(config.instance_load_time * 2) # let's wait a little longer
          try:
            # try to build again
            self.inst[i] = (factory.builder.build(self, self.offset))
          except Exception, e:
            if getattr(e, 'typeName', None) == 'com.sun.star.connection.NoConnectException':
              # try to kill OOo instance and restart again
              Log.debug('no connection - restarting OOo')
              start.killInstance(i)
              start.startInstance(i, keep_control=True)
              time.sleep(config.instance_load_time)
              try:
                self.inst[i] = (factory.builder.build(self, self.offset))
              except Exception, e:
                if getattr(e, 'typeName', None) == 'com.sun.star.connection.NoConnectException':
                  # no connection
                  Log.debug('no connection - will retry in %d' % (config.instance_load_time * 2))
                  time.sleep(config.instance_load_time * 2) # let's wait a little longer
                  # try for the last time
                  self.inst[i] = (factory.builder.build(self, self.offset))
                else:
                  raise
            else:
              raise
        else:
          raise
    except Exception, e:
      Log.logException(e)
      raise

pool = Pool()

# vim: shiftwidth=2
