#!/usr/bin/python
##############################################################################
#
# Copyright (c) 2002, 2006 Nexedi SARL and Contributors. All Rights Reserved.
#                    Bartlomiej Gorny <bartek@erp5.pl>
#                    Kevin Deldycke <kevin@nexedi.com>
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsability of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# garantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

import os
import re
import sys
import threading
from urllib import quote
# Add oood home as a place to look for import
sys.path.append(os.path.abspath(os.path.dirname(__file__)))

# Standard python libs
import getopt, time, glob
from commands import getstatusoutput
# ood python libs
import config
from logger import Log

WIN = False
if os.name == 'nt':
    WIN = True
    from subprocess import Popen, PIPE
    import winservice as winutils

"""
  Size of the pool is defined in config under the 'pool_size' parameter.

  User directories user_0, user_1 etc. will be created automatically
    by OOo on startup in the ~/tmp folder of the running user.

  TODO: Merge start.py and runserw.py file in one file to manage oood. This
        file should be called "ooodctl.py", ala zopectl.
"""


def _l(msg, instance_id, level=0):
  """
    This method put instance related messages in the log
  """
  LOG_HEAD = "Instance #%s - " % instance_id
  Log.info('%s%s' % (LOG_HEAD, msg))


def startInstance(i, on_top=False, keep_control=False):
  """
    We spawn a new process and record its pid to be able to kill it if necessary.
  """
  _l("Starting...", i)
  instance_port = config.pool_port_range_start + i
  if WIN:
    BIN = "soffice.exe"
  else:
    BIN = "soffice"
  cmd = '%s/%s' % (config.prog_dir, BIN)
  drive, tail = os.path.splitdrive(config.run_dir.strip('/'))
  run_dir = drive +  quote(tail)
  instance_dir = '/%s/oood_instance_%d' %(run_dir,i)

  # copy Common.xcu to skip registration dialog
  common_xcu_dir = '%s/user/registry/data/org/openoffice/Office' % instance_dir
  try:
    os.makedirs(common_xcu_dir)
  except OSError:
    pass
  open('%s/Common.xcu' % common_xcu_dir, 'w').write(
    open(os.path.join(os.path.dirname(__file__), 'Common.xcu')).read())

  # "'-display :%d' % config.virtual_display_id" argument is not working, that's why we use
  # environnment variable. See "man xhost" for more details.
  new_context = os.environ
  if on_top:
    args_and_env = [ BIN  # First arg is always the name of the binary software we execute
                 , '-accept=socket,host=%s,port=%d;urp;' % (config.pool_host, instance_port)
                 , '-env:UserInstallation=file://%s' % instance_dir
                 , new_context
                 ]
  else:
    new_context['DISPLAY'] = ':%s' % config.virtual_display_id
    args_and_env = [ BIN  # First arg is always the name of the binary software we execute
                 , '-invisible'
                 , '-nologo'
                 , '-nodefault'
                 , '-norestore'
                 , '-nofirststartwizard'
                 , '-accept=socket,host=%s,port=%d;urp;' % (config.pool_host, instance_port)
                 , '-env:UserInstallation=file://%s' % instance_dir
                 , new_context
                 ]
  if WIN:
    cmd_and_args = [cmd] + args_and_env[1:-1]
    process = Popen(cmd_and_args)
    pid = process.pid
  else:
    pid = os.spawnlpe(os.P_NOWAIT, cmd, *args_and_env)
  _l("Listening at %s:%s" % (config.pool_host, instance_port), i)
  pidfile = os.path.join(config.run_dir, 'instance_%d.pid' % i)
  open(pidfile, 'w').write(str(pid))
  if keep_control:
      # if we start instance from running oood (because we did kill/restart),
      # we have to call "wait" on it to keep it under control
      # otherwise when it is killed it becomes a zombie
      # so we establish a waiting thread, not to block the whole process
      _l("Starting waiting thread for process %s" % pid, i)
      if WIN:
        waiting_thread = threading.Thread(target=justWait, args=(process))
      else:
        waiting_thread = threading.Thread(target=justWait, args=(pid,))
      waiting_thread.start()
  _l("Started as process  %s" % pid, i)

def justWait(pid):
    if WIN:
      pid.wait()
    else:
      os.waitpid(pid, 0)

def killInstance(i):
  _l("Kill requested", i)
  pid_file_path = os.path.join(config.run_dir, 'instance_%s.pid' % i)
  if os.path.exists(pid_file_path):
    pid_file = open(pid_file_path, 'r')
    instance_master_pid = int(pid_file.read())
    pid_file.close()
    # OOo instance span on several process, find all of them
    # on windows this is not necessary because the supprocesses
    # are automatically terminated when we kill the OOo parent process
    if WIN:
        instance_pid_list = winutils.getChildrenOfPid(instance_master_pid)
        if instance_pid_list:
            instance_pid_list.append(instance_master_pid)
    else:
        pid_list = os.popen('ps -A -o pid,ppid').read().split('\n')
        instance_pid_list = []
        for pid_line in pid_list:
          pid_pair = re.findall('(\d+)', pid_line)
          if len(pid_pair) == 2:
            pid_pair = [int(pid) for pid in pid_pair]
            if instance_master_pid in pid_pair:
              instance_pid_list.append(pid_pair[0])
    if len(instance_pid_list) > 0:
      _l("Is still running and span on several processes: %s" % ', '.join([str(x) for x in instance_pid_list]), i)
      # Kill all instance processes
      for pid in instance_pid_list:
        _l("Killing process %s..." % pid, i)
        if WIN:
            winutils.SafeTerminateProcess(pid)
        else:
            os.kill(pid, 9)
      _l("Killed", i)
    else:
      _l("Was not properly shutdown", i)
    os.remove(pid_file_path)
    _l("Cleaned", i)
  else:
    _l("Didn't exist, so no need to kill", i)


def initPool(on_top=False):
  """
    This mehod initialize the pool: it kill all previous OOo instances (if any) and re-populate the pool.
  """
  Log.info("Pool - Initialize: populate the pool with %s instances" % config.pool_size)
  flushPool()
  for i in range(config.pool_size):
    startInstance(i, on_top)
    Log.info("Pool - Wait %s seconds to let the instance #%s start" % (config.instance_load_time, i))
    time.sleep(config.instance_load_time)
  Log.info("Pool - Initialized: %s instances started" % config.pool_size)


def flushPool():
  """
    This method flush the pool: it kill all OOo instances to empty the pool.
  """
  Log.info("Pool - Flush requested: kill all instances")
  # pool_size can change dynamiccaly, so clear every running OOo intance
  max_pool_size = config.pool_size
  for pid_file in glob.glob('%s*' % config.run_dir):
    file_name = os.path.basename(pid_file)
    if file_name.startswith('instance_') and file_name.endswith('.pid'):
      instance_id = int(file_name[len('instance_'):len(file_name)-len('.pid')])
      old_pool_size = instance_id + 1
      if old_pool_size > max_pool_size:
        max_pool_size = old_pool_size
  for i in range(max_pool_size):
    killInstance(i)
  Log.info("Pool - Flushed: all instances killed")


def startVirtualFrameBuffer():
  """
    This method start a virtual frame buffer (Xvfb)
  """
  VIRTUAL_SCREEN = 0
  XORG_TMP_PATH  = "/tmp/.X11-unix"
  import pwd
  oood_uid = os.getuid()
  oood_gid = os.getgid()
  # Clean previous virtual display if found
  killVirtualFrameBuffer()
  # Remove bad previous socket
  xorg_socket = os.path.join(XORG_TMP_PATH, 'X%s' % config.virtual_display_id)
  if os.path.exists(xorg_socket) and \
    (os.stat(xorg_socket).st_uid != oood_uid or \
     os.stat(xorg_socket).st_gid != oood_gid):
    msg = "Pool - Can't start Virtual Frame Buffer because '%s' is not owned by %s" \
               % (xorg_socket,pwd.getpwuid(os.getuid())[0] )
    Log.warning(msg)
    print msg
    sys.exit(1)
  # Start the new virtual display server
  os.system("Xvfb -ac :%s -screen %s 800x600x16 -fbdir %s &" % ( config.virtual_display_id
                                                               , VIRTUAL_SCREEN
                                                               , config.run_dir
                                                               ))
  Log.info("Pool - Virtual frame buffer started")
  Log.info("Pool - To see the virtual display, run 'xwud -in %s/Xvfb_screen%s'" % (config.run_dir, VIRTUAL_SCREEN))


def killVirtualFrameBuffer():
  """
    This method kill the virtual frame buffer (Xvfb)
  """
  # Search all Xvfb process running on display ID given in the config file
  result = getstatusoutput("ps ax | grep 'Xvfb' | grep ':%s'" % config.virtual_display_id)
  # Search the running process
  killed = False
  for process in result[1].splitlines():
    if process.find('grep') == -1:
      pid = int(process.strip().split(' ')[0])
      os.kill(pid, 9)
      killed = True
  if killed:
    Log.info("Pool - Virtual frame buffer killed")


def showThreadingStatus():
  import xmlrpclib
  uri = 'http://%s:%d' % (config.server_host, config.server_port)
  sp = xmlrpclib.ServerProxy(uri)
  try:
      print sp.threadingStatus()
  except Exception, e:
      print 'we got exception:'
      print e
      print 'probably server is not running'


def showPoolStatus():
  import xmlrpclib
  uri = 'http://%s:%d' % (config.server_host, config.server_port)
  sp = xmlrpclib.ServerProxy(uri)
  try:
      print sp.poolStatus()
  except Exception, e:
      print 'we got exception:'
      print e
      print 'probably server is not running'


#def showPoolStatus():
  #"""
    #Show Pool details and statistics.
  #"""
  ## TODO: must be done (!)
  #print "===== Pool Statistics ====="
  #print "Status: %s" % "TODO"
  #print "Size: %s" % config.pool_size
  #print "Uptime: %s" % "TODO"
  #print "OOo instance running: %s" % "TODO"
  ##print "OOo instance stopped: %s" % "TODO"
  #print "OOo instance auto-restarted since launch: %s" % "TODO"


def usage():
  print """Usage: %s [options]
  -i, --init
      Initialize the pool: kill all previous OOo instances (if any) and re-populate the pool.
  -t, --top
      Start on top
  -f, --flush
      Flush the pool: kill all OOo instances to empty the pool.
  -s, --status, --stat
      Show pool statistics and status.
  -r, --threads, --th
      Show currently running threads
  -h, --help
      Show this screen.
""" % sys.argv[0]


if __name__=='__main__':

  # Parse options and parameters
  try:
    opts, args = getopt.getopt( sys.argv[1:]
                              , "hifsdtr"
                              , ["help", "init", "flush", "status", "stat", "threads", "top"]
                              )
    # Check args number
    if len(opts) == 0:
      raise getopt.GetoptError('', '')
  except getopt.GetoptError:
    # Print help information and exit on command line error
    usage()
    sys.exit(2)

  # Start action according parameters
  for o, a in opts:
    if o in ("-h", "--help"):
      usage()
      sys.exit(0)
    elif o in ("-i", "--init"):
      print "You are using new implementation of oood; no need to run start.py --init"
      sys.exit(0)
    elif o in ("-t", "--top"):
      initPool(on_top=True)
      sys.exit(0)
    elif o in ("-f", "--flush"):
      flushPool()
      if not WIN:
        killVirtualFrameBuffer()
      sys.exit(0)
    elif o in ("-s", "--status", "--stat"):
      showPoolStatus()
      print """
      This screen tells you about status and statistics of the oood server:
      - how long has it been running
      - how many times it was restarted ("emergency restart")
      - how long has it been running since the last restart
      - how many requests has it processed
      Also, there is a some information about a worker that the oood is using:
      - how long has it been running
      - how many requests has it processed
      - how long (in seconds) has it been processing the current file (if this value reaches
        the instance_timeout set in config, the worker is killed and restarted)
      - the (stripped) name of the file it is now processing
      Sometimes the whole row is filled with "xxxx" marks - it means that this worker is being restarted, or that
      it has died and this slot in the pool is empty, so the server is running a smaller pool. If the pool 
      size goes to zero, the whole server is restarted.
      """
      sys.exit(0)
    elif o in ("-r", "--threads", "--th"):
      showThreadingStatus()
      print """
      What you should see above is a list of currently running threads:
      MainThread is the main server thread.
      There is a number of threads which have no function or args, and which do not change;
      their number should be equal to the number of OOo instances running, they are waiting for
      OOo instances to terminate (die, crash, be killed) so that they do not become zombies.
      There is also a threads whose id increases every two seconds - these are created
      in response to the script you are currently looking at.
      Other threads are being launched, started, stopped and terminated by dispatcher to wait for 
      certain operations to complete, and to launch instance restart of emergency restart of the whole
      server if these operations time out.
      """
      sys.exit(0)

  # Exit on general error
  sys.exit(1)
