#!/usr/bin/python
# vim: set fileencoding=iso-8859-1 :
##############################################################################
#
# Copyright (c) 2002, 2006 Nexedi SARL and Contributors. All Rights Reserved.
#                    Bartlomiej Gorny <bartek@erp5.pl>
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsability of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# garantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

"""
  Test various oood operations - getting and setting metadata, converting basic file types
  to ODF, generating basic file formats
"""

import sys, base64, unittest
import getopt
import re
from xmlrpclib import *
import config
from oood_common import responseFactory, Request

enc = base64.encodestring

opts, args = getopt.getopt(sys.argv[1:], '', ["offset=", "port=", "host="])

port = None
host = None

offset_str = ''
offset = 0

for o, a in opts:
  if o == '--offset':
    offset_str = a
    offset = int(a)
  elif o == '--port':
    port = int(a)
  elif o == '--host':
    host = a

doc_dir = 'doc%s' % offset_str

if not port:
  port = config.server_port + offset
  print port
if not host:
  host = config.server_host
  print host

try:
  f = open('%s/test.odt' % doc_dir, 'rb')
except IOError:
  print "you need a %s subdir with appropriate documents" % doc_dir
  sys.exit(1)

sp = ServerProxy('http://%s:%d' % (host, port))

def extractExtensions(item_list):
  return [item[0] for item in item_list]

class TestMetaOperations(unittest.TestCase):

  def testSetGetMeta(self):
    # setting metadata
    mime = 'application/vnd.oasis.opendocument.text'
    newmeta = {'title':'HELLOOOOOO', 'reference':'blaaaaaah', 'MIMEType':mime}
    data = open(doc_dir+'/test.odt','rb').read()
    res = sp.setmetadata('test.odt', base64.encodestring(data), newmeta)
    res = responseFactory(res)
    self.failUnless(res.isOk())
    open(doc_dir+'/out/test_changed.odt', 'wb').write(res.data)
    # getting metadata back after the change
    data = open(doc_dir+'/out/test_changed.odt','rb').read()
    res = sp.getmetadata('test_changed.odt', base64.encodestring(data))
    res = responseFactory(res)
    self.failUnless(res.isOk())
    self.assertEqual(res.meta['title'], 'HELLOOOOOO')
    self.assertEqual(res.meta['reference'], 'blaaaaaah')

class TestFileOperations(unittest.TestCase):

  def testConvertDoc(self):
    """
      This test passes arguments the old way.
    """
    data = open(doc_dir+'/test.doc', 'rb').read()
    res = sp.convert('test.doc', base64.encodestring(data))
    res = responseFactory(res)
    self.failUnless(res.isOk())
    self.assertEqual(res.mime, 'application/vnd.oasis.opendocument.text')
    open(doc_dir+'/out/test.odt', 'wb').write(res.data)
    res = sp.getAllowedTargetItemList(res.mime)
    res = responseFactory(res)
    self.failUnless(res.isOk())
    target_list = res.response_data
    self.failUnless('doc' in extractExtensions(target_list))

  def testConvertXls(self):
    """
      This test uses Request object with constructor.
    """
    data = open(doc_dir+'/test.xls', 'rb').read()
    request = Request(filename='text.xls', data=data)
    res = sp.convert(request)
    res = responseFactory(res)
    self.failUnless(res.isOk())
    self.assertEqual(res.mime, 'application/vnd.oasis.opendocument.spreadsheet')
    open(doc_dir+'/out/test.ods', 'wb').write(res.data)
    request = Request(mimetype=res.mime)
    res = sp.getAllowedTargetItemList(request)
    res = responseFactory(res)
    self.failUnless(res.isOk())
    target_list = res.response_data
    self.failUnless('csv' in extractExtensions(target_list))

  def testConvertPpt(self):
    """
      This test uses Request object setting attributes.
    """
    data = open(doc_dir+'/test.ppt', 'rb').read()
    request = Request()
    request.filename = 'test.ppt'
    request.data = data
    res = sp.convert(request)
    res = responseFactory(res)
    self.failUnless(res.isOk())
    self.assertEqual(res.mime,'application/vnd.oasis.opendocument.presentation')
    open(doc_dir+'/out/test.odp', 'wb').write(res.data)
    request = Request()
    request.mimetype = res.mime
    res = sp.getAllowedTargetItemList(request)
    res = responseFactory(res)
    self.failUnless(res.isOk())
    target_list = res.response_data
    self.failUnless('impr.pdf' in extractExtensions(target_list))

  def testPdfTextGeneration(self):
    data = open(doc_dir+'/test.odt', 'rb').read()
    request = Request(filename='test.odt', data=data, extension='pdf')
    res = sp.generate(request)
    self.assert_(res)
    res = responseFactory(res)
    self.failUnless(res.isOk())
    self.assertEqual(res.mime, 'application/pdf')
    open(doc_dir+'/out/test.pdf', 'wb').write(res.data)

  def testPdfTextGenerationWithIllegalChars(self):
    """
      This tries to break xmlrpc protocol by putting illegal character in filename
      (the Request class should fix it before proceeding).
    """
    data = open(doc_dir+'/test.odt', 'rb').read()
    request = Request(filename='tyciy.odt', data=data, extension='pdf')
    res = sp.generate(request)
    self.assert_(res)

  def testPdfTextGenerationWithNoFilename(self):
    """
      If filename was not given in the request, the server should produce its own
      from random digits.
    """
    data = open(doc_dir+'/test.odt', 'rb').read()
    request = Request(data=data, extension='pdf')
    res = sp.generate(request)
    self.assert_(res)
    res = responseFactory(res)
    self.failUnless(res.isOk())
    self.failUnless(re.match('.*\d{2}$', res.filename))

  def testPdfCalcGeneration(self):
    data = open(doc_dir+'/test.ods', 'rb').read()
    res = sp.generate('test.ods', enc(data), -1, 'calc.pdf')
    self.assert_(res)
    res = responseFactory(res)
    self.failUnless(res.isOk())
    self.assertEqual(res.mime, 'application/pdf')
    open(doc_dir+'/out/test.calc.pdf', 'wb').write(res.data)

  def testPdfImpressGeneration(self):
    data = open(doc_dir+'/test.odp', 'rb').read()
    res = sp.generate('test.odp', enc(data), -1, 'impr.pdf')
    self.assert_(res)
    res = responseFactory(res)
    self.failUnless(res.isOk())
    self.assertEqual(res.mime, 'application/pdf')
    open(doc_dir+'/out/test.impr.pdf', 'wb').write(res.data)

  def testHtmlWriterGeneration(self):
    self.generateFile('odt', 'writer.html')

  def testRtfGeneration(self):
    self.generateFile('odt', 'rtf')

  def testPptGeneration(self):
    self.generateFile('odp', 'ppt')

  def testDocGeneration(self):
    self.generateFile('odt', 'doc')

  def testTxtGeneration(self):
    self.generateFile('odt', 'txt')

  def testHtmlCalcGeneration(self):
    self.generateFile('ods', 'calc.html')

  def testCsvGeneration(self):
    self.generateFile('ods', 'csv')

  def testJpgGeneration(self):
    self.generateFile('odg', 'jpg')

  def testDrawPdfGeneration(self):
    self.generateFile('odg', 'draw.pdf')

  def testTxtFromDocGeneration(self):
    self.generateFile('doc', 'txt')

  def testDocFromTxtGeneration(self):
    data = open(doc_dir+'/test.iso.txt', 'rb').read()
    request = Request(filename='test.iso.txt', data=data, extension='doc')
    res = sp.generate(request)
    self.assert_(res)
    res = responseFactory(res)
    self.failUnless(res.isOk())
    open(doc_dir+'/out/test.iso.doc', 'wb').write(res.data)

  def generateFile(self,src,ext):
    data = open(doc_dir+'/test.%s' % src, 'rb').read()
    request = Request(filename='test.%s' % src, data=data, extension=ext)
    res = sp.generate(request)
    self.assert_(res)
    res = responseFactory(res)
    self.failUnless(res.isOk())
    open(doc_dir+'/out/test.%s' % ext, 'wb').write(res.data)

if __name__=='__main__':
  tests=(TestMetaOperations, TestFileOperations)
  for t in tests:
    suite = unittest.makeSuite(t)
    unittest.TextTestRunner(verbosity=2).run(suite)




# vim: filetype=python syntax=python shiftwidth=2 
