#!/usr/bin/python
##############################################################################
#
# Copyright (c) 2002, 2006 Nexedi SARL and Contributors. All Rights Reserved.
#                    Bartlomiej Gorny <bartek@erp5.pl>
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

"""
  This test sends random numbers of random ODF files to oood, to convert them to random formats,
  all at random intervals. This is to simulate production environment.
  First it creates a list of all ODF docs which are in the system to pick up files from it.
  Fails if server raises exception returns something which is not expected type, unless it is
  an integer which means an error code - the serwer had a problem and handled it properly.
"""

import getopt
import os
import random
import sys
import threading
import time
import unittest
from xmlrpclib import *

import config
from oood_common import responseFactory, code2message, Request

### configuration
max_batch_size = 5
max_interval = 15
###

opts, args = getopt.getopt(sys.argv[1:], '', ["offset=", "port=", "host=", 'total-conversions='])

port = None
host = None
total_conversions = 100

offset_str = ''
offset = 0

for o, a in opts:
  if o == '--offset':
    offset_str = a
    offset = int(a)
  elif o == '--port':
    port = int(a)
  elif o == '--host':
    host = a
  elif o == '--total-conversions':
    total_conversions = int(a)

doc_dir = 'doc%s' % offset_str

if not port:
  port = config.server_port + offset
  print port
if not host:
  host = config.server_host
  print host

try:
  f = open('%s/test.odt' % doc_dir)
except IOError:
  print "you need a %s subdir with appropriate documents" % doc_dir
  sys.exit(1)

sp = ServerProxy('http://%s:%d' % (host, port))

# create a list of od? files if not present
# first check if the oood_home is writeable
if not os.access(config.oood_home, os.W_OK):
  print 'the %s directory is not writeable for me' % config.oood_home
  print 'I cant proceed'
  print 'please make it writeable so that I can create "all_odf_docs" file'
  sys.exit(1)

all_odfs_name = config.oood_home + '/all_odf_docs'
all_odfs_creation_command = 'find /home | egrep "\.od[stp]{1}$" > %s' % all_odfs_name
if not os.path.exists(all_odfs_name):
  print "creating a list of ODF documents in the system - this can take a while, but I do it only once..."
  os.system(all_odfs_creation_command)
file_names = open(all_odfs_name).readlines()
if len(file_names) == 0:
  print 'all_odf_docs file is empty'
  print 'please remove it and start again'
  sys.exit(1)
#file_names= file_names[50:]

def log(msg):
  """
    a helper function to print and log the same thing
  """
  stamp = time.strftime('%d, %H:%M:%S')
  thread_name = threading.currentThread().getName()
  msg = '#%s %s : %s : %s' % (offset_str, stamp, thread_name, msg)
  f = open('test.log', 'a')
  f.write(msg + '\n')
  f.close()
  print(msg)

ext2mime = {
    'odt':'application/vnd.oasis.opendocument.text',
    'ods':'application/vnd.oasis.opendocument.spreadsheet',
    'odp':'application/vnd.oasis.opendocument.presentation'
  }

lock = threading.Lock()

class TestScalability(unittest.TestCase):

  result_dict = {}
  start = time.time()

  def logResult(self, res):
    self.result_dict[res] = self.result_dict.get(res, 0) + 1
    f = open('stats.txt', 'w')
    uptime = int(time.time() - self.start)
    days = uptime / (24 * 3600)
    uptime = uptime - days * (24 * 3600)
    hours = uptime / 3600
    uptime = uptime - hours * 3600
    minutes = uptime / 60
    uptime = uptime - minutes * 60
    seconds = uptime
    f.write('last update: %s\n' % time.strftime('%Y-%m-%d %H:%M:%S'))
    f.write('test running for %s days, %s hours, %s minutes, %s seconds\n' % (days, hours, minutes, seconds))
    f.write('test configuration: max_batch_size %s, max_interval %s, total_conversions %s\n' % (max_batch_size, max_interval, total_conversions))
    f.write('oood configuration: pool_size %s, instance_load_time %s, instance_timeout %s\n' % (config.pool_size, config.instance_load_time, config.instance_timeout))
    f.write('\n')
    for res, count in self.result_dict.items():
      f.write('%s : %s\n' % (count, res))
    f.write('==========\ntotal: %s\n' % sum(self.result_dict.values()))
    f.write('to go: %s\n' % (total_conversions - sum(self.result_dict.values())))
    f.write('\n')
    f.write("""
    Server responses:
      200: 'OK',
      401: 'requested method is not provided',
      402: 'the document could not be processed',
      403: 'timeout while processing document',
      404: 'malformed request',
      501: 'the server is restarting',
      502: 'pool is empty, the server will be restarted',
      503: 'the server is out of sync and will be restarted',
      504: 'unknown error'
    """)
    f.close()

  def generateFormat(self):
    """
      take a random file from od? file list, randomly choose one target format
      from among allowed targets, generate, check if it returned
    """
    local = threading.local()
    # make sure the file is there
    while True:
      local.fname = random.choice(file_names).strip()
      if os.path.exists(local.fname):
        break
    ext = local.fname[-3:]
    local.request = Request(extension=ext)
    local.response = responseFactory(sp.getAllowedTargetItemList(local.request))
    if local.response.code != 200:
      log('got error code: %s "%s" (getAllowedTargetItemList for %s)' % (local.response.code, code2message(local.response.code), ext))
      self.logResult(local.response.code)
      return
    local.target_list = local.response.response_data
    local.tgt = random.choice(local.target_list)
    local.data = open(local.fname).read()
    local.request = Request(filename=local.fname, data=local.data, extension=local.tgt[0])
    local.response = sp.generate(local.request)
    local.response = responseFactory(local.response)
    if local.response.code != 200:
      log('xxx got error code: %s "%s" (%s --> %s)' % (local.response.code, code2message(local.response.code), local.fname, local.tgt[1]))
      self.logResult(local.response.code)
      return
    log('--- got result: %s (%s --> %s)' % (local.response.code, local.fname, local.tgt[1]))
    self.logResult(local.response.code)
    self.failUnless(local.response.code == 200)

  def testRandomLoad(self):
    """
      random number of generation requests (1-5)fired up at the same time
      by separate threads
      sleep random number of seconds (1-30)
      do it 100 times
    """
    counter = 0
    #repeat = 2
    print ' '
    #for iteration in range(repeat):
    while True:
      random.seed()
      batch_size = random.randint(1, max_batch_size)
      interval = random.randint(1, max_interval)
      log('batch %d (total %d), interval %d' % (batch_size, counter + batch_size, interval))
      for i in range(batch_size):
        counter += 1
        t = threading.Thread(target=self.generateFormat)
        t.start()
        if counter == total_conversions:
          return
      time.sleep(interval)

if __name__=='__main__':
  tests=(TestScalability,)
  for t in tests:
    suite=unittest.makeSuite(t)
    unittest.TextTestRunner(verbosity=2).run(suite)
