#!/usr/bin/python
##############################################################################
#
# Copyright (c) 2002, 2006 Nexedi SARL and Contributors. All Rights Reserved.
#                    Bartlomiej Gorny <bartek@erp5.pl>
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsability of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# garantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

import os, sys, time, string
import config
sys.path.insert(0, config.uno_path)
import uno
from unohelper import systemPathToFileUrl, fileUrlToSystemPath,absolutize
from com.sun.star.beans import PropertyValue
from com.sun.star.beans import PropertyExistException
from com.sun.star.task import ErrorCodeIOException
from com.sun.star.lang import IllegalArgumentException

import lib
from mimemapper import mimemapper
from logger import Log

def caseUp(s):
  """
    OOo uses erp5-like naming convention, so we do the same
    except for MIMEType
  """
  if s == 'MIMEType':return s
  return ''.join([st.capitalize() for st in s.split('_')])

def caseDown(s):
  try:
    s = s.encode('utf-8', 'ignore')
  except AttributeError:
    pass
  if len(s) <= 1:return s.lower()
  if s == 'MIMEType':return s
  r = lambda c: c in string.uppercase and '_' + c or c
  s=[s[0]] + map(r, s[1:])
  s = ''.join(s).lower()
  return s

class Worker(object):

  """
    This class encapsulates an OOo instance providing interface
    to functions. Can be accessed directly by functions
    "convert" etc., but should be through wrappers
    "run_convert", which is safe and resets the worker finally.
  """

  # for automatic interface generation
  public_funcs=('convert', 'generate', 'setmetadata', 'getmetadata', 'getmetadatatoolong',
      'generate_with_user_fields')
  inProps = PropertyValue( "Hidden" , 0 , True, 0 ),

  metafields=('title', 'subject', 'keywords', 'description', 'MIMEType')
  userfields=('reference', 'version', 'language')
  # this attribute says if the worker is doing something, or waiting for request
  busy = False
  # this attribute says if the worker is removed from the pool because of a timeout
  dead = False

  def __init__(self,idx,desktop,pool = None):
    self.idx = idx
    self.desktop = desktop
    self.pool = pool
    self.cwd = systemPathToFileUrl(config.oood_home)
    self._reset()
    self._generateFuncs()
    self.creation_time = int(time.time())
    self.start_time = 0
    self.counter = 0

  def __setattr__(self, name, value):
    """
      When the worker is flagged as busy, we start measuring
      its "busytime"
    """
    if name == 'busy':
      if value:
        self.start_time = int(time.time())
      else:
        self.start_time = 0
    object.__setattr__(self, name, value)

  def __getattribute__(self, name):
    """
      Used to measure uptime and busytime
    """
    current_time = int(time.time())
    if name == 'uptime':
      return current_time - self.creation_time
    if name == 'busytime':
      if self.start_time == 0:
        return 0
      return current_time - self.start_time
    return object.__getattribute__(self, name)

  def convert(self, kw):
    """
      convertion to OOo, produces metadata and OOo file url
    """
    self._loadFile(kw['filename'], kw)
    self._convert(kw)
    # if the doc doesn't have title or reference, we use filename for them
    # only on ERP5 object (we don't set them on file)
    # strip filename of #_ prefix
    try:
      name = os.path.basename(kw['filename'])
      strippedfilename = name[name.index('_') + 1:]
    except ValueError:
      strippedfilename = name
    if not self.metadata.get('title'):
      self.metadata['title'] = strippedfilename
    if not self.metadata.get('reference'):
      self.metadata['reference'] = strippedfilename
    kw['mime'] = self.metadata.get('MIMEType')
    kw['meta'] = self.metadata
    kw['newfilename'] = fileUrlToSystemPath(self.destUrl)

  def generate(self, kw):
    """
      generate file in a requested format, save file url
      and new mimetype in kw dictionary
    """
    if mimemapper.isOdf(extension=kw['extension']):
      return self.convert(kw)
    self._loadFile(kw['filename'], kw)
    self._generate(kw['extension'])
    mime = mimemapper.getMimeFor(kw['extension'])
    kw['mime'] = mime
    kw['newfilename'] = fileUrlToSystemPath(self.destUrl)

  """ MasterFields have different names in 2.2 and 2.4 """

  OPENOFFICE_24_PREFIX = "com.sun.star.text.fieldmaster.User.%s"
  OPENOFFICE_22_PREFIX = "com.sun.star.text.FieldMaster.User.%s"
  
  def _replace_key(self, key, prefix, data, masters):
    actual_key = prefix % key
    if actual_key in masters.getElementNames():
        try:
            master = masters.getByName(actual_key)
            master.setPropertyValue('Content', unicode(data[key]))
        except:
            pass
      
  def _fill_user_fields(self, user_fields):
      masters = self.doc.getTextFieldMasters()
      available_keys = masters.getElementNames()
      for key in user_fields.keys():
          self._replace_key(key, self.OPENOFFICE_24_PREFIX, user_fields, masters)
          self._replace_key(key, self.OPENOFFICE_22_PREFIX, user_fields, masters)      

  def generate_with_user_fields(self, kw):
      """
          behaves like generate; however, it assumes that in the kw
          there is a key named user_fields, which is used to
          update MasterFields inside the provided document;
      """
      self._loadFile(kw['filename'], kw)
      self._fill_user_fields(kw['user_fields'])
      self._generate(kw['extension'])
      mime = mimemapper.getMimeFor(kw['extension'])
      kw['mime'] = mime[0] # it is a tuple; can be multiple but we don't know which exactly was requested
      kw['newfilename'] = fileUrlToSystemPath(self.destUrl)

  def setmetadata(self, kw):
    """
      set metadata on OOo file (fname does not change)
    """
    self._loadFile(kw['filename'])
    self._setMetadata(kw['meta'])
    kw['newfilename'] = kw['filename']
    self._saveFile()
    return True

  def getmetadatatoolong(self,kw):
    """
      for debugging - to see if timeout works
    """
    self._loadFile(kw['filename'])
    kw['meta'] = self._getMetadata()
    time.sleep(config.instance_timeout + 120)
    return True
  
  def getmetadata(self, kw):
    """
      Get metadata from the OOo file
    """
    self._loadFile(kw['filename'])
    meta = self._getMetadata()
    kw['meta']=meta
    return meta

  def printDocument(self, kw):
    """
      print the document directly on local printer, without any conversion
    """
    self._loadFile(kw['filename'], kw)
    self._printDocument()
   
  def _generateFuncs(self):
    """
      generates functions run_<funcname>
      XXX we should probably change naming convention and generate
      automatically based on name pattern
      or use decorators
    """
    for f in self.public_funcs:
      fname = 'run_' + f
      setattr(self, fname,self._gener(f))

  def _gener(self, f):
    """
      Wrapper for API functions
      runs the function, and then resets the worker
      no matter what happened
    """
    def _run(*args, **kwargs):
      """
        if anything goes wrong, we reset the worker
        and log exception
      """
      try:
        try:
          res = func(*args, **kwargs)
        except Exception, e:
          Log.logException(e)
          raise
      finally:
        self._reset()
      return res
    func = getattr(self, f)
    return _run

  def _reset(self):
    """
      reset all worker's properties
    """
    Log.info('[worker] resetting worker %d' % self.idx)
    try:
      if hasattr(self, 'doc') and self.doc:
        self.doc.close(True)
    except Exception, e:
      # we log exception but do not reraise
      Log.logException(e)
    for atr in ('fileUrl','doc','dinfo','metadata','destUrl'):
      setattr(self,atr,None)

  def _loadFile(self,fname, kw={}):
    """
      Loads a document from local filesystem, using auto-generated file name.
      Uses default OOo auto-detection mechanism.
      Client can enforce use of text (encoded) filter by explicitly specifying encoding.
      XXX TODO: the same for CSV
    """
    Log.info('[worker] trying to load %s' % fname)
    inProps = self.inProps
    if kw.has_key('encoding'):
      prop_name = PropertyValue("FilterName", 0, "Text (encoded)", 0)
      prop_encoding = PropertyValue("FilterOptions", 0, kw['encoding'], 0)
      inProps = (self.inProps[0], prop_name, prop_encoding)
    if 'orig_format' in kw:
      filter_name = mimemapper.getInputFilterForMime(kw['orig_format'])
      if filter_name is not None:
        prop_name = PropertyValue('FilterName', 0, filter_name, 0)
        inProps = (self.inProps[0], prop_name)
    self.fileUrl = absolutize(self.cwd,systemPathToFileUrl(fname))
    try:
      self.doc = self.desktop.loadComponentFromURL(self.fileUrl, '_blank', 0, inProps)
      if getattr(self.doc, 'getTypes', None) is not None:
          if 'com.sun.star.util.XRefreshable' in [i.typeName for i in self.doc.getTypes()]:
              self.doc.refresh()
    except IllegalArgumentException:
      Log.error('[worker] ' + fname + ' could not be loaded into OpenOffice')
      raise lib.NotLoaded(fname + ' could not be loaded into OpenOffice')
    if not self.doc:
      Log.error('[worker] ' + fname + ' was not loaded into OpenOffice')
      raise lib.NotLoaded(self.fileUrl)
    self.counter += 1
    self.metadata = self._getMetadata()

  def _saveFile(self):
    self.doc.store()

  def _closeFile(self):
    self.doc.close(True)

  def _generate(self, format):
    """
      From the ODF file that was submitted to the worker, now generate an output file
      in a desired format
    """
    export_filter = mimemapper.getFilterForExt(format)
    type_info = mimemapper.getTypePropertyDict(extension=format)
    if type_info is None:
      raise lib.IllegalFormat(format)
    self.destUrl = self.fileUrl + '.' + format # we want it very unique
    outprops = [PropertyValue('FilterName', 0, export_filter, 0),
                PropertyValue('Overwrite', 0, True, 0),]
    export_props = type_info.get('export_props', None)
    if export_props is not None:
      outprops.extend(export_props)
    try:
      self.doc.storeToURL(self.destUrl, tuple(outprops))
    except ErrorCodeIOException:
      raise lib.GenerationFailed(self.fileUrl + ' to format ' + format)

  def _getMetadata(self):
    """
      extract metadata, assign as dictionary
      no MIME check - we get them also before conversion
      we convert names between two naming conventions
    """
    self.dinfo = self.doc.getDocumentInfo()
    metadata = {}
    for n in self.metafields:
      v = self.dinfo.getPropertyValue(caseUp(n))
      if v:
        metadata[n] = v
    for i in range(self.dinfo.getUserFieldCount()):
      n = self.dinfo.getUserFieldName(i)
      v = self.dinfo.getUserFieldValue(i)
      if caseDown(n) in self.userfields:
        if v:
          metadata[caseDown(n)] = v
    metadata['FilterName'] = [x.Value for x in self.doc.getArgs() if x.Name == 'FilterName'][0]
    return metadata

  def _convert(self, kw):
    """
      This method converts non-ODF format into an ODF
      only known mimetypes,store as a file with
      appropriate extension 
      we append extension instead of
      replacing, because filename may have unusual extension or
      none at all
    """
    odf_format_info = mimemapper.getOdfInfo(filter_name=self.metadata['FilterName'])
    if odf_format_info is None:
      raise lib.IllegalMimeType(self.metadata['MIMEType'])
    ext = odf_format_info[0]
    self.destUrl = self.fileUrl + '.' + ext
    self.doc.storeAsURL(self.destUrl, ())
    # we change mimetype to OOo before sending metadata back
    self.metadata['MIMEType'] = odf_format_info[1]

  def _setMetadata(self, meta):
    """
      set metadata given as a dictionary
      add new property if does not exist
      OOo uses titlecased property names
    """
    if not mimemapper.isOdf(filter_name=self.metadata['FilterName']):
      raise Exception('can not set metadata on non-ODF file')
    self.dinfo = self.doc.getDocumentInfo()
    userfieldcounter = 0
    maxcount = self.dinfo.getUserFieldCount()
    for k,v in meta.items():
      K = caseUp(k)
      if k in self.metafields:
        try:
          self.dinfo.addProperty(K, 0, v)
        except PropertyExistException:
          self.dinfo.setPropertyValue(K, v)
      elif k in self.userfields:
        if userfieldcounter < maxcount:
          self.dinfo.setUserFieldName(userfieldcounter, K)
          self.dinfo.setUserFieldValue(userfieldcounter, v)
          userfieldcounter += 1
          # we silently ignore if there is too many user fields
      else:
        raise lib.IllegalMetaAttribute(k)

  def _printDocument(self):
    """
      prints out current document on the printer queue set in config
    """
    try:
      prter = self.doc.getPrinter()
      print_method = getattr(self.doc, 'print')
      prter[0].Value = config.printer_name
      self.doc.setPrinter(prter)
      print_method(())
    except Exception, e:
      Log.logException(e)
      raise e

  def getAllowedTargetItemList(self, request):
    """
      list types which can be generated from given type
      Type can be given as:
      - mimetype
      - filename extension
      - document type ('text', 'spreadsheet', 'presentation' or 'drawing')
    """
    document_type = request.get('document_type', None)
    extension = request.get('extension', None)
    # hack to accept deprecated arguments
    mimetype = request.get('mimetype') or request.get('filename')
    return mimemapper.getAllowedTo(mimetype=mimetype, document_type=document_type, extension=extension)


if __name__ == '__main__':
  print caseDown('Reference')
  #w = Worker(None)
  #print w.__dict__
  #w.run_convert('s')
#
# vim: shiftwidth=2
